@@ -4,11 +4,14 @@
  * 1990-09-04  V 1.1.0
  * 2000-05-31  V 1.3.0 beta
  * 2000-10-18  V 1.3.0 final
+ * 2002-01-16  V 1.3.1
+ * 2004-01-09  V 1.3.2
+ * 2013-08-23  V 1.4.0
  *
  * NOTE: Edit this file with tabstop=4 !
  *
- * Copyright 1990-2000 by Gerhard Buergmann
- * Gerhard.Buergmann@altavista.net
+ * Copyright 1990-2013 by Gerhard Buergmann
+ * gerhard@puon.at
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -30,22 +33,28 @@
 #	include <locale.h>
 #endif
 
-#ifdef	__MSDOS__
+#if defined(__MSDOS__) && !defined(DJGPP)
 #	define PRINTF	cprintf
 #else
 #	define PRINTF	printf
 #ifndef HELPFILE
+#  ifdef DJGPP
+#	define HELPFILE "/dev/env/DJDIR/lib/bmore.help"
+#  else
 #	define HELPFILE "/usr/local/lib/bmore.help"
+#  endif
 #endif
 #endif
 
 #include "bmore.h"
 
-char	*copyright  = "Copyright (C) 1990-2000 by Gerhard Buergmann";
+char	*copyright  = "Copyright (C) 1990-2013 by Gerhard Buergmann";
 
 int		maxx, maxy;
+int		mymaxx = 0, mymaxy = 0;
 char	*name = NULL;
-char	sstring[MAXCMD] = "";
+char	sstring[MAXCMD] = "";	/* string for search */
+char	estring[MAXCMD] = "";	/* string for shell escape */
 char	string[MAXCMD];
 FILE	*curr_file = NULL, *help_file;
 int		AnzAdd;
@@ -60,11 +69,13 @@ char	numarr[64];		/* string for collecting number */
 char	addr_form[15];
 
 int		ascii_flag = 0;
-int		d_flag = 0;
+int		dup_print_flag = 0;
+int		c_flag = 0, d_flag = 0;
+int		exval = 0;
 int		init_search = 0;
 char	buffer1[MAXCMD], buffer2[MAXCMD];
 int		out_len;
-int		corr = 0, to_print;
+int		corr = 0, do_header = 0, to_print;
 off_t	init_byte = 0;
 off_t	last_search = 0;
 off_t	screen_home, filesize;
@@ -72,24 +83,28 @@ off_t	bytepos, oldpos;
 int		prompt = 1;
 char	helppath[MAXCMD];
 
-static	char	*progname;
+static	char	progname[10];
 static	char	cmdbuf[MAXCMD];
 static	int		cnt = 0;
 static	int		icnt = 0;
 static	int		smode;
 
-char	search_pat[BUFFER];	/* / or ? command */
+/* char	search_pat[BUFFER];	*/  /* / or ? command */
+char	bmore_search_pat[BUFFER];   /* / or ? command */
 char	*emptyclass = "Empty byte class '[]' or '[^]'";
 
 
 /* -a   ASCII mode
  * -d   beginners mode
+ * -c	clear before displaying
  * -i   ignore case
+ * -n	number of lines/screen
+ * -w   width of screen
  */
 void
 usage()
 {
-	fprintf(stderr, "Usage: %s [-adi] [+linenum | +/pattern] name1 name2 ...\n", progname);
+	fprintf(stderr, "Usage: %s [-acdi] [-lines] [+linenum | +/pattern] name1 name2 ...\n", progname);
 	exit(1);
 }
 
@@ -106,15 +121,14 @@ main(argc, argv)
 	long	count;
 	int		i, n = 1;
 	int		d_line, r_line, z_line;
-
-
-#ifdef __MSDOS__
 	char	*poi;
 
+
+#if defined(__MSDOS__) && !defined(DJGPP)
 	strcpy(helppath, argv[0]);
 	poi = strrchr(helppath, '\\');
 	*poi = '\0';
-	strcat(helppath, "\\MORE.HLP");
+	strcat(helppath, "\\BMORE.HLP");
 #else
 	strncpy(helppath, HELPFILE, MAXCMD - 1);
 #endif
@@ -123,23 +137,46 @@ main(argc, argv)
 	setlocale(LC_ALL, "");
 #endif
 
-	progname = argv[0];
+	poi = strrchr(argv[0], DELIM);
+ 
+	if (poi) strncpy(progname, ++poi, 9);
+	else strncpy(progname, argv[0], 9);
+	strtok(progname, ".");
 
 	while (n < argc) {
 		switch (argv[n][0]) {
 		case '-':
-			i = 1;
-			while (argv[n][i] != '\0') {
-				switch (argv[n][i]) {
-				case 'a':	ascii_flag++;
-							break;
-				case 'd':	d_flag++;
-							break;
-				case 'i':	ignore_case++;
-							break;
-				default:	usage();
+			if (argv[n][1] >= '0' && argv[n][1] <= '9') {
+				sscanf(&argv[n][1], "%dx%d", &mymaxy, &mymaxx);
+			} else if (argv[n][1] == 'n') {
+				if (argv[n+1] == NULL || argv[n+1][0] == '-') {
+					usage();
+				} else {
+					sscanf(&argv[++n][0], "%d", &mymaxy);
+				}
+			} else if (argv[n][1] == 'w') {
+				if (argv[n+1] == NULL || argv[n+1][0] == '-') {
+					usage();
+				} else {
+					sscanf(&argv[++n][0], "%d", &mymaxx);
+				}
+			} else {
+				i = 1;
+				while (argv[n][i] != '\0') {
+					switch (argv[n][i]) {
+					case 'a':	ascii_flag++;
+								break;
+					case 'c':	c_flag++;
+								break;
+					case 'd':	d_flag++;
+								break;
+					case 'i':	ignore_case++;
+								break;
+					default:		
+								usage();
+					}
+					i++;
 				}
-				i++;
 			}
 			n++;
 			break;
@@ -148,11 +185,7 @@ main(argc, argv)
 				init_search = argv[n][1];
 				strcpy(sstring, &argv[n][2]);
 			} else {
-				if (argv[n][1] == '0') {
-					init_byte = (off_t)strtol(argv[n] + 1, NULL, 16);
-				} else {
-					init_byte = (off_t)strtol(argv[n] + 1, NULL, 10);
-				}
+				init_byte = strtoll(argv[n] + 1, NULL, 0);
 			}
 			n++;
 			break;
@@ -164,48 +197,79 @@ main(argc, argv)
 			break;
 		}
 	}
-	if (numfiles == 0) {
-		curr_file = stdin;
-		if (isatty(fileno(stdin)) != 0) usage();
-	} else {
-		open_file(name);
-		file_nr = 1;
-		fseek(curr_file, init_byte, SEEK_SET);
-		bytepos += init_byte;
-	}
-	screen_home = bytepos;
-
 	initterm();
 	set_tty();
 	maxy -= 2;
+	if (mymaxy) {
+		maxy = mymaxy;
+	}
 	z_line = maxy;
 	d_line = maxy / 2;
 	r_line = 1;
 
+	if (numfiles == 0) {
+		curr_file = stdin;
+		if (isatty(fileno(stdin)) != 0) {
+			reset_tty();
+			usage();
+		}
+	} else {
+		file_nr = 1;
+		while (open_file(name)) {
+			do_next(1);
+		}
+		if (exval) {
+		} else {
+			fseeko(curr_file, init_byte, SEEK_SET);
+			bytepos += init_byte;
+		}
+	}
+	screen_home = bytepos;
+
 	AnzAdd = 10;
 	strcpy(addr_form,  "%08lX  ");
 
 	if (ascii_flag)
-		 out_len = ((maxx - AnzAdd - 1) / 4) * 4;
+		out_len = ((maxx - AnzAdd - 1) / 4) * 4;
 	else
 		out_len = ((maxx - AnzAdd - 1) / 16) * 4;
+	if (mymaxx) {
+		out_len = mymaxx;
+	}
 
 	if (init_search)
 		bmsearch(init_search);
 
 	if (no_tty) {
-		while(!printout(1));
+		int	fileloop;
+
+		for (fileloop = 0; fileloop < numfiles; fileloop++) {
+			while(!printout(1));
+			do_next(1);
+			open_file(name);
+		}
+		if (curr_file) fclose(curr_file);
 		reset_tty();
-		exit(0);
+		exit(exval);
 	}
-	if (printout(maxy)) {
-		do_next(1);
+	if (!exval) {
+		if (printout(maxy)) {
+			do_next(1);
+		}
 	}
+	signal(SIGINT, sig);
+	signal(SIGQUIT, sig);
 	/* main loop */
 	do {
 		to_print = 0;
+		dup_print_flag = 0;
 		if (prompt) {
-			highvideo();
+			if (prompt == 2) {
+				while (open_file(name)) {
+					do_next(1);
+				}
+			}
+			highlight();
 			PRINTF("--More--");
 			if (prompt == 2) {
 				PRINTF("(Next file: %s)", name);
@@ -214,13 +278,15 @@ main(argc, argv)
 			}
 
 			if (d_flag) PRINTF("[Press space to continue, 'q' to quit]");
-			normvideo();
+			normal();
 			fflush(stdout);
 		}
 		ch = vgetc();
+		/*
 		if (prompt == 2) {
 			open_file(name);
 		}
+		*/
 		prompt = 1;
 		PRINTF("\r");
 		while (ch >= '0' && ch <= '9') {
@@ -245,23 +311,27 @@ main(argc, argv)
 
 		switch (ch) {
 		case ' ':	/*  Display next k lines of text [current screen size] */
+					dup_print_flag = 1;
 					if (precount > 0) to_print = precount;
 						else to_print = maxy;
 					break;
 		case 'z':	/* Display next k lines of bytes [current screen size]* */
+					dup_print_flag = 1;
 					if (precount > 0) z_line = precount;
 					to_print = z_line;
 					break;
 		case '\r':
 		case '\n':	/* Display next k lines of text [current screen size]* */
+					dup_print_flag = 1;
 					if (precount > 0) r_line = precount;
 					to_print = r_line;
 					break;
 		case 'q':
 		case 'Q':
-					clreol();
+					cleartoeol();
+					fclose(curr_file);
 					reset_tty();
-					exit(0);
+					exit(exval);
 		case ':' :	
 					switch (colon) {
 					case 'f':
@@ -278,7 +348,7 @@ main(argc, argv)
 						if (precount < 1) precount = 1;
 						do_next(precount);
 						PRINTF("\r");
-						clreol();
+						cleartoeol();
 						PRINTF("\n...Skipping to file %s\r\n\r\n", name);
 						prompt = 2;
 						break;
@@ -286,46 +356,55 @@ main(argc, argv)
 						if (precount < 1) precount = 1;
 						do_next(-precount);
 						PRINTF("\r");
-						clreol();
+						cleartoeol();
 						PRINTF("\n...Skipping back to file %s\r\n\r\n", name);
 						prompt = 2;
 						break;
 					case 'q':
-						clreol();
+						cleartoeol();
+						fclose(curr_file);
 						reset_tty();
-						exit(0);
+						exit(exval);
 						break;
 					case '!':
 						if (!no_intty) {
-							clreol();
-							if (rdline(ch)) break;
-							doshell(sstring);
+							cleartoeol();
+							if (rdline(colon, estring)) break;
+							doshell(estring);
 							PRINTF("------------------------\r\n");
 							break;
 						}
 					default:
-						beep();
+						bmbeep();
 					}
 					break;
+		case '!':
+			if (!no_intty) {
+				cleartoeol();
+				if (rdline(ch, estring)) break;
+				doshell(estring);
+				PRINTF("------------------------\r\n");
+				break;
+			}
 		case 'd':	/* Scroll k lines [current scroll size, initially 11]* */
-		case CTRL('D'):
+		case BVICTRL('D'):
 					if (precount > 0) d_line = precount;
 					to_print = d_line;
 					break;
-		case CTRL('L'):   	/*** REDRAW SCREEN ***/
+		case BVICTRL('L'):   	/*** REDRAW SCREEN ***/
 					if (no_intty) {
-						beep();
+						bmbeep();
 					} else {
-						clrscr();
+						clearscreen();
 						to_print = maxy + 1;
-						fseek(curr_file, screen_home, SEEK_SET);
+						fseeko(curr_file, screen_home, SEEK_SET);
 						bytepos = screen_home;
 					}
 					break;
 		case 'b':		/* Skip backwards k screenfuls of text [1] */
-		case CTRL('B'):
+		case BVICTRL('B'):
 					if (no_intty) {
-						beep();
+						bmbeep();
 					} else {
 						if (precount < 1) precount = 1;
 						PRINTF("...back %ld page", precount);
@@ -336,7 +415,7 @@ main(argc, argv)
 						}
 						screen_home -= (maxy + 1) * out_len;
 						if (screen_home < 0) screen_home = 0;
-						fseek(curr_file, screen_home, SEEK_SET);
+						fseeko(curr_file, screen_home, SEEK_SET);
 						bytepos = screen_home;
 						to_print = maxy + 1;
 					}
@@ -350,7 +429,7 @@ main(argc, argv)
 						count = precount;
 					}
 					putchar('\r');
-					clreol();
+					cleartoeol();
 					PRINTF("\n...skipping %ld line", count);
 					if (count > 1) {
 						PRINTF("s\r\n\r\n");
@@ -358,18 +437,19 @@ main(argc, argv)
 						PRINTF("\r\n\r\n");
 					}
 					screen_home += (count + maxy) * out_len;
-					fseek(curr_file, screen_home, SEEK_SET);
+					fseeko(curr_file, screen_home, SEEK_SET);
 					bytepos = screen_home;
 					to_print = maxy;
 					break;
 		case '\\':  
 					if (ascii_flag) {
-						beep();
+						bmbeep();
 						break;
 					}
 		case '/':	/**** Search String ****/
 					if (!repeat) {
-						if (rdline(ch)) break;
+						cleartoeol();
+						if (rdline(ch, sstring)) break;
 					}
 		case 'n': 		/**** Search Next ****/
 		case 'N':   
@@ -380,20 +460,20 @@ main(argc, argv)
 					break;
 		case '\'':   
 					if (no_intty) {
-						beep();
+						bmbeep();
 					} else {
 						bytepos = last_search;
-						fseek(curr_file, bytepos, SEEK_SET);
+						fseeko(curr_file, bytepos, SEEK_SET);
 						screen_home = bytepos;
 						to_print = maxy;
 						PRINTF("\r");
-						clreol();
+						cleartoeol();
 						PRINTF("\n\r\n***Back***\r\n\r\n");
 					}
 					break;
 		case '=':
 					prompt = 0;
-					clreol();
+					cleartoeol();
 					PRINTF("%lX hex  %lu dec", (unsigned long)bytepos,
 						(unsigned long)bytepos);
 					fflush(stdout);
@@ -412,7 +492,7 @@ main(argc, argv)
 		case 'w':
 		case 'v':
 					if (!no_intty) {
-						clreol();
+						cleartoeol();
 						if (ch == 'v') {
 							sprintf(string, "bvi +%lu %s", 
 								(unsigned long)(screen_home + 
@@ -431,7 +511,7 @@ main(argc, argv)
 					if (d_flag) {
 						emsg("[Press 'h' for instructions.]");
 					} else {
-						beep();
+						bmbeep();
 					}
 					break;
 		}
@@ -445,13 +525,18 @@ main(argc, argv)
 
 
 int
-rdline(ch)
-	int	ch;
+rdline(ch, sstring)
+	int		ch;
+	char	*sstring;
 {
-	int	i = 0;
-	int	ch1 = 0;
-	
-	clreol();
+	int		i = 0;
+	int		ch1 = 0;
+	char	bstring[MAXCMD];
+
+	if (ch == '!') {
+		strcpy(bstring, sstring);
+		sstring[0] = '\0';
+	}
 	putchar(ch);
 	fflush(stdout);
 
@@ -463,11 +548,20 @@ rdline(ch)
 			if (i) {
 				sstring[--i] = '\0';
 				PRINTF("\r%c%s", ch, sstring);
-				clreol();
+				cleartoeol();
 			} else {
 				ch1 = ESC;
 				break;
 			}
+		} else if (ch1 == '!' && i == 0) {
+			if (bstring[0] == '\0') {
+				emsg("No previous command");
+				return 1;
+			}
+			putchar(ch1);
+			PRINTF("\r%c%s", ch, bstring);
+			strcat(sstring, bstring);
+			i = strlen(sstring);
 		} else {
 			putchar(ch1);
 			sstring[i++] = ch1;
@@ -476,7 +570,7 @@ rdline(ch)
 	}
 	if (ch1 == ESC) {
 		putchar('\r');
-		clreol();
+		cleartoeol();
 		return 1;
 	}
 	if (i) sstring[i] = '\0';
@@ -490,8 +584,9 @@ do_next(n)
 {
 	if (numfiles) {
 		if (n == 1 && file_nr == numfiles) {
+			if (curr_file) fclose(curr_file);
 			reset_tty();
-			exit(0);
+			exit(exval);
 		}
 		if ((file_nr + n) > numfiles)
 			file_nr = numfiles;
@@ -503,13 +598,14 @@ do_next(n)
 		free(name);
 		name = strdup(*(files + file_nr - 1));
 	} else {
+		if (curr_file) fclose(curr_file);
 		reset_tty();
-		exit(0);
+		exit(exval);
 	}
 }
 
 
-void
+int
 open_file(name)
 	char *name;
 {
@@ -518,19 +614,15 @@ open_file(name)
 	if (stat(name, &buf) > -1) {
 		filesize = buf.st_size;
 	}
-	if (numfiles > 1) {
-		PRINTF("\r");
-		clreol();
-		PRINTF("\n::::::::::::::\r\n%s\r\n::::::::::::::\r\n", name);
-		corr = 2;
-	}
-	if (curr_file != NULL) fclose(curr_file);
+	if (curr_file) fclose(curr_file);
+	if (numfiles > 1) do_header = 1;
 	if ((curr_file = fopen(name, "rb")) == NULL) {
-		reset_tty();
 		perror(name);
-		exit(1);
+		exval = 1;
+		return 1;
 	}
 	bytepos = screen_home = 0;
+	return 0;
 }
 
 
@@ -542,7 +634,7 @@ putline(buf, num)
 	int			print_pos;
 	unsigned	char	ch;
 
-	PRINTF(addr_form, bytepos);
+	PRINTF(addr_form, (unsigned long)bytepos);
 	for (print_pos = 0; print_pos < num; print_pos++) {
 		ch = buf[print_pos];
 		if (!ascii_flag) {
@@ -554,8 +646,16 @@ putline(buf, num)
 		else
 			*(string + print_pos) = '.';
 	}
+	for (; print_pos < out_len; print_pos++) {
+		if (!ascii_flag) {
+		    PRINTF("   ");
+		}
+		++bytepos;
+		*(string + print_pos) = ' ';
+	}
 	*(string + num) = '\0';
-	PRINTF("%s\r\n", string);
+	if (no_tty) PRINTF("%s\n", string);
+	else PRINTF("%s\r\n", string);
 }
 
 
@@ -567,6 +667,20 @@ printout(lns)
 	int			doub = 0;
 	static		int		flag;
 	
+	if (c_flag) {
+		clearscreen();
+	}
+	if (do_header) {
+		if (no_tty) {
+			PRINTF("::::::::::::::\n%s\n::::::::::::::\n", name);
+		} else {
+			PRINTF("\r");
+			cleartoeol();
+			PRINTF("::::::::::::::\r\n%s\r\n::::::::::::::\r\n", name);
+		}
+		do_header = 0;
+		corr = 2;
+	}
 	if (corr && (lns > maxy - 2)) lns -= corr;
 	corr = 0;
 	do {
@@ -575,14 +689,14 @@ printout(lns)
 			buffer1[num] = c;
 		}
 		if (!num) return 1;
-		if (memcmp(buffer1, buffer2, num) || !bytepos ) {
+		if (memcmp(buffer1, buffer2, num) || !bytepos || !dup_print_flag) {
 			memcpy(buffer2, buffer1, num);
 			putline(buffer2, num);
 			if (!no_tty) flag = TRUE;
 			lns--;
 		} else {
 			if (flag) {
-				clreol();
+				cleartoeol();
 				PRINTF("*\r\n");
 				lns--;
 			} else {
@@ -596,6 +710,7 @@ printout(lns)
 			if (screen_home < 0) screen_home = 0;
 			return 0;
 		}
+		dup_print_flag = 1;
 	} while(num);
 	return 1;
 }
@@ -774,16 +889,19 @@ bmsearch(ch)
 		return;
 	}
 	if (ch == '/') {
-		if (ascii_comp(search_pat, sstring)) return;
+		/* if (ascii_comp(search_pat, sstring)) return; */
+		if (ascii_comp(bmore_search_pat, sstring)) return;
 	}
 	if (ch == '\\') {
-		if (hex_comp(search_pat, sstring)) return;
+		/* if (hex_comp(search_pat, sstring)) return; */
+		if (hex_comp(bmore_search_pat, sstring)) return;
 	}
 	oldpos = bytepos;
 	last_search = screen_home;
 	if (precount < 1) precount = 1;
 	while (precount--) {
-		while ((i = bmregexec(search_pat)) == 0);
+		/* while ((i = bmregexec(search_pat)) == 0); */
+		while ((i = bmregexec(bmore_search_pat)) == 0);
 		if (i == 1) {
 			screen_home = bytepos;
 			to_print = maxy;
@@ -792,13 +910,13 @@ bmsearch(ch)
 				PRINTF("\r\nPattern not found\r\n");
 				do_next(1);
 			} else {
+/*
 sprintf(string, "Pattern not found %d - %ul", i, (unsigned long)bytepos);
 emsg(string);
-/*
-				emsg("Pattern not found");
 */
+				emsg("Pattern not found");
 				bytepos = oldpos;
-				fseek(curr_file, bytepos, SEEK_SET);
+				fseeko(curr_file, bytepos, SEEK_SET);
 				break;
 			}
 		}
@@ -818,10 +936,16 @@ emsg(s)
 	char	*s;
 {
 	putchar('\r');
-	clreol();
-	highvideo();
-	PRINTF(s);
-	normvideo();
+	cleartoeol();
+	highlight();
+	PRINTF("%s", s);
+	normal();
 	fflush(stdout);
 	prompt = 0;
 }
+
+
+void
+bmbeep() {
+	putchar(7);
+}
